// Copyright 2024 the V8 project authors. All rights reserved.
// Use of this source code is governed by a BSD-style license that can be
// found in the LICENSE file.

#ifndef V8_OBJECTS_JS_DISPOSABLE_STACK_H_
#define V8_OBJECTS_JS_DISPOSABLE_STACK_H_

#include "src/base/bit-field.h"
#include "src/handles/handles.h"
#include "src/objects/js-objects.h"
#include "torque-generated/bit-fields.h"

// Has to be the last include (doesn't have include guards):
#include "src/objects/object-macros.h"

namespace v8 {
namespace internal {

#include "torque-generated/src/objects/js-disposable-stack-tq.inc"

// Valid states for a DisposableStack.
// https://arai-a.github.io/ecma262-compare/?pr=3000&id=sec-disposablestack-objects
enum class DisposableStackState { kDisposed, kPending };

// kValueIsReceiver: Call the method with no argument
// kValueIsArgument: Pass the value as the argument to the dispose method,
// `disposablestack.prototype.adopt` is the only method that uses
// kValueIsArgument as DisposeMethodCallType.
enum class DisposeMethodCallType { kValueIsReceiver = 0, kValueIsArgument = 1 };

class JSDisposableStack
    : public TorqueGeneratedJSDisposableStack<JSDisposableStack, JSObject> {
 public:
  DECL_PRINTER(JSDisposableStack)
  DECL_VERIFIER(JSDisposableStack)

  DEFINE_TORQUE_GENERATED_DISPOSABLE_STACK_STATUS()
  inline DisposableStackState state() const;
  inline void set_state(DisposableStackState value);
  DECL_INT_ACCESSORS(length)

  static void Initialize(Isolate* isolate,
                         DirectHandle<JSDisposableStack> stack);
  static void Add(Isolate* isolate,
                  DirectHandle<JSDisposableStack> disposable_stack,
                  DirectHandle<Object> value, DirectHandle<Object> method,
                  DisposeMethodCallType type);
  static MaybeHandle<Object> CheckValueAndGetDisposeMethod(
      Isolate* isolate, Handle<Object> value);
  static Maybe<bool> DisposeResources(
      Isolate* isolate, DirectHandle<JSDisposableStack> disposable_stack,
      MaybeHandle<Object> maybe_error);

  TQ_OBJECT_CONSTRUCTORS(JSDisposableStack)
};

}  // namespace internal
}  // namespace v8

#include "src/objects/object-macros-undef.h"

#endif  // V8_OBJECTS_JS_DISPOSABLE_STACK_H_
